import asyncio
import gc
import logging
import os
import tracemalloc

import pamqp
import pytest

from async_generator import yield_, async_generator
from yarl import URL

import aiormq
from aiormq import Connection


POLICIES = [asyncio.DefaultEventLoopPolicy]
POLICY_IDS = ['asyncio']

try:
    import uvloop
    POLICIES.append(uvloop.EventLoopPolicy)
    POLICY_IDS.append('uvloop')
except ImportError:
    pass


@pytest.fixture(params=POLICIES, ids=POLICY_IDS)
def event_loop(request):
    policy = request.param()

    asyncio.get_event_loop().close()
    asyncio.set_event_loop_policy(policy)

    loop = asyncio.new_event_loop()     # type: asyncio.AbstractEventLoop
    loop.set_debug(True)
    asyncio.set_event_loop(loop)

    original = asyncio.get_event_loop

    def getter_mock():
        raise RuntimeError('asyncio.get_event_loop() call forbidden')

    asyncio.get_event_loop = getter_mock

    nocatch_marker = request.node.get_closest_marker(
        'no_catch_loop_exceptions'
    )

    exceptions = list()
    if not nocatch_marker:
        loop.set_exception_handler(lambda l, c: exceptions.append(c))

    try:
        yield loop

        if exceptions:
            raise RuntimeError(exceptions)

    finally:
        asyncio.get_event_loop = original
        asyncio.set_event_loop_policy(None)
        del policy

        loop.close()
        del loop


def cert_path(*args):
    return os.path.join(
        os.path.abspath(os.path.dirname(__file__)), 'certs', *args
    )


amqp_urls = [
    URL(os.getenv('AMQP_URL', 'amqp://guest:guest@localhost/')),
    URL(os.getenv(
        'AMQP_URL', 'amqp://guest:guest@localhost/'
    )).with_scheme('amqps').with_query({
        "cafile": cert_path('ca.pem'),
        "no_verify_ssl": 1,
    }),
    URL(os.getenv(
        'AMQP_URL', 'amqp://guest:guest@localhost/'
    )).with_scheme('amqps').with_query({
        "cafile": cert_path('ca.pem'),
        "keyfile": cert_path('client.key'),
        "certfile": cert_path('client.pem'),
        "no_verify_ssl": 1,
    })
]


@pytest.fixture(params=amqp_urls, ids=['amqp', 'amqps', 'amqps-client'])
@async_generator
async def amqp_connection(request, event_loop):
    connection = Connection(request.param, loop=event_loop)

    await connection.connect()

    try:
        await yield_(connection)
    finally:
        await connection.close()


channel_params = [
    dict(channel_number=None, frame_buffer=10, publisher_confirms=True),
    dict(channel_number=None, frame_buffer=1, publisher_confirms=True),
    dict(channel_number=None, frame_buffer=10, publisher_confirms=False),
    dict(channel_number=None, frame_buffer=1, publisher_confirms=False),
]


@pytest.fixture(params=channel_params)
@async_generator
async def amqp_channel(request, amqp_connection):
    try:
        await yield_(await amqp_connection.channel(**request.param))
    finally:
        await amqp_connection.close()


skip_when_quick_test = pytest.mark.skipif(
    os.getenv("TEST_QUICK") is not None,
    reason='quick test'
)


@pytest.fixture(autouse=True)
def memory_tracer():
    tracemalloc.start()
    tracemalloc.clear_traces()

    filters = (
        tracemalloc.Filter(True, aiormq.__file__),
        tracemalloc.Filter(True, pamqp.__file__),
        tracemalloc.Filter(True, asyncio.__file__),
    )

    snapshot_before = tracemalloc.take_snapshot().filter_traces(filters)

    def format_stat(stats):
        items = [
            "TOP STATS:",
            "%-90s %6s %6s %6s" % ("Traceback", "line", "size", "count")
        ]

        for stat in stats:
            fname = stat.traceback[0].filename
            lineno = stat.traceback[0].lineno
            items.append(
                "%-90s %6s %6s %6s" % (
                    fname,
                    lineno,
                    stat.size_diff,
                    stat.count_diff
                )
            )

        return "\n".join(items)

    try:
        yield

        gc.collect()

        snapshot_after = tracemalloc.take_snapshot().filter_traces(filters)

        top_stats = snapshot_after.compare_to(
            snapshot_before, 'lineno', cumulative=True
        )

        if top_stats:
            logging.error(format_stat(top_stats))
            raise AssertionError("Possible memory leak")
    finally:
        tracemalloc.stop()



import asyncio
import uuid

import pytest

import aiormq


pytestmark = pytest.mark.asyncio


async def test_simple(amqp_channel: aiormq.Channel):
    await amqp_channel.basic_qos(prefetch_count=1)
    assert amqp_channel.number

    queue = asyncio.Queue()

    deaclare_ok = await amqp_channel.queue_declare(auto_delete=True)
    consume_ok = await amqp_channel.basic_consume(deaclare_ok.queue, queue.put)
    await amqp_channel.basic_publish(
        b'foo', routing_key=deaclare_ok.queue,
        properties=aiormq.spec.Basic.Properties(message_id='123')
    )

    message = await queue.get()     # type: DeliveredMessage
    assert message.body == b'foo'

    cancel_ok = await amqp_channel.basic_cancel(consume_ok.consumer_tag)
    assert cancel_ok.consumer_tag == consume_ok.consumer_tag
    await amqp_channel.queue_delete(deaclare_ok.queue)

    deaclare_ok = await amqp_channel.queue_declare(auto_delete=True)
    await amqp_channel.basic_publish(b'foo bar', routing_key=deaclare_ok.queue)

    message = await amqp_channel.basic_get(deaclare_ok.queue, no_ack=True)
    assert message.body == b'foo bar'


async def test_blank_body(amqp_channel: aiormq.Channel):
    await amqp_channel.basic_qos(prefetch_count=1)
    assert amqp_channel.number

    queue = asyncio.Queue()

    deaclare_ok = await amqp_channel.queue_declare(auto_delete=True)
    consume_ok = await amqp_channel.basic_consume(deaclare_ok.queue, queue.put)
    await amqp_channel.basic_publish(
        b'', routing_key=deaclare_ok.queue,
        properties=aiormq.spec.Basic.Properties(message_id='123')
    )

    message = await queue.get()     # type: DeliveredMessage
    assert message.body == b''

    cancel_ok = await amqp_channel.basic_cancel(consume_ok.consumer_tag)
    assert cancel_ok.consumer_tag == consume_ok.consumer_tag
    await amqp_channel.queue_delete(deaclare_ok.queue)

    deaclare_ok = await amqp_channel.queue_declare(auto_delete=True)
    await amqp_channel.basic_publish(b'foo bar', routing_key=deaclare_ok.queue)

    message = await amqp_channel.basic_get(deaclare_ok.queue, no_ack=True)
    assert message.body == b'foo bar'


@pytest.mark.no_catch_loop_exceptions
async def test_bad_consumer(amqp_channel: aiormq.Channel, event_loop):
    channel = amqp_channel      # type: aiormq.Channel
    await channel.basic_qos(prefetch_count=1)

    declare_ok = await channel.queue_declare()

    future = event_loop.create_future()

    await channel.basic_publish(b'urgent', routing_key=declare_ok.queue)

    consumer_tag = event_loop.create_future()

    async def bad_consumer(message):
        await channel.basic_cancel(await consumer_tag)
        future.set_result(message)
        raise Exception

    consume_ok = await channel.basic_consume(
        declare_ok.queue, bad_consumer, no_ack=False
    )

    consumer_tag.set_result(consume_ok.consumer_tag)

    message = await future
    await channel.basic_reject(message.delivery.delivery_tag, requeue=True)
    assert message.body == b'urgent'

    future = event_loop.create_future()

    await channel.basic_consume(
        declare_ok.queue, future.set_result, no_ack=True
    )

    message = await future

    assert message.body == b'urgent'
    await channel.basic_ack(message.delivery.delivery_tag)


async def test_ack_nack_reject(amqp_channel: aiormq.Channel, event_loop):
    channel = amqp_channel                      # type: aiormq.Channel
    await channel.basic_qos(prefetch_count=1)

    declare_ok = await channel.queue_declare(auto_delete=True)
    queue = asyncio.Queue(loop=event_loop)

    await channel.basic_consume(declare_ok.queue, queue.put, no_ack=False)

    await channel.basic_publish(b'rejected', routing_key=declare_ok.queue)
    message = await queue.get()
    assert message.body == b'rejected'
    await channel.basic_reject(message.delivery.delivery_tag, requeue=False)

    await channel.basic_publish(b'nacked', routing_key=declare_ok.queue)
    message = await queue.get()
    assert message.body == b'nacked'
    await channel.basic_nack(message.delivery.delivery_tag, requeue=False)

    await channel.basic_publish(b'acked', routing_key=declare_ok.queue)
    message = await queue.get()
    assert message.body == b'acked'
    await channel.basic_ack(message.delivery.delivery_tag)


async def test_confirm_multiple(amqp_channel: aiormq.Channel):
    """
    RabbitMQ has been observed to send confirmations in a strange pattern
    when publishing simultaneously where only some messages are delivered
    to a queue. It sends acks like this 1 2 4 5(multiple, confirming also 3).
    This test is probably inconsequential without publisher_confirms
    This is a regression for https://github.com/mosquito/aiormq/issues/10
    """
    channel = amqp_channel                      # type: aiormq.Channel
    exchange = uuid.uuid4().hex
    await channel.exchange_declare(exchange, exchange_type='topic')
    try:
        declare_ok = await channel.queue_declare(exclusive=True)
        await channel.queue_bind(declare_ok.queue, exchange,
                                 routing_key='test.5')

        for i in range(10):
            messages = [channel.basic_publish(b'test', exchange=exchange,
                                              routing_key='test.{}'.format(i))
                        for i in range(10)]
            _, pending = await asyncio.wait(messages, timeout=0.2)
            assert not pending, 'not all publishes were completed (confirmed)'
            await asyncio.sleep(0.05)
    finally:
        await channel.exchange_delete(exchange)


async def test_exclusive_queue_locked(amqp_connection):
    channel0 = await amqp_connection.channel()
    channel1 = await amqp_connection.channel()

    qname = str(uuid.uuid4())

    await channel0.queue_declare(qname, exclusive=True)

    try:
        await channel0.basic_consume(qname, print, exclusive=True)

        with pytest.raises(aiormq.exceptions.ChannelLockedResource):
            await channel1.queue_declare(qname)
            await channel1.basic_consume(qname, print, exclusive=True)
    finally:
        await channel0.queue_delete(qname)



import asyncio

import pytest

from aiormq.base import FutureStore


@pytest.fixture
def root_store(event_loop):
    store = FutureStore(loop=event_loop)
    try:
        yield store
    finally:
        event_loop.run_until_complete(
            store.reject_all(Exception("Cancelling"))
        )


@pytest.fixture
def child_store(event_loop, root_store):
    store = root_store.get_child()
    try:
        yield store
    finally:
        event_loop.run_until_complete(
            store.reject_all(Exception("Cancelling"))
        )


@pytest.mark.asyncio
async def test_reject_all(event_loop, root_store: FutureStore,
                          child_store: FutureStore):

    future1 = root_store.create_future()
    future2 = child_store.create_future()

    assert root_store.futures
    assert child_store.futures

    await root_store.reject_all(RuntimeError)
    await asyncio.sleep(0.1, loop=event_loop)

    assert isinstance(future1.exception(), RuntimeError)
    assert isinstance(future2.exception(), RuntimeError)
    assert not root_store.futures
    assert not child_store.futures


@pytest.mark.asyncio
async def test_result(event_loop, root_store: FutureStore,
                      child_store: FutureStore):

    async def result():
        await asyncio.sleep(0.1, loop=event_loop)
        return 'result'

    assert await child_store.create_task(result()) == 'result'


@pytest.mark.asyncio
async def test_siblings(event_loop, root_store: FutureStore,
                        child_store: FutureStore):

    async def coro(store):
        await asyncio.sleep(0.1, loop=event_loop)
        await store.reject_all(RuntimeError)

    task1 = child_store.create_task(coro(child_store))
    assert root_store.futures
    assert child_store.futures

    with pytest.raises(RuntimeError):
        await task1

    await asyncio.sleep(0.1, loop=event_loop)

    assert not root_store.futures
    assert not child_store.futures

    child = child_store.get_child().get_child().get_child()
    task = child.create_task(coro(child))

    assert root_store.futures
    assert child_store.futures
    assert child.futures

    with pytest.raises(RuntimeError):
        await task

    await asyncio.sleep(0.1, loop=event_loop)

    assert not root_store.futures
    assert not child_store.futures
    assert not child.futures



import asyncio
import logging
import platform
import ssl
import typing
from base64 import b64decode
from contextlib import suppress

import pamqp.frame
from pamqp import ProtocolHeader
from pamqp import specification as spec
from pamqp.heartbeat import Heartbeat
from yarl import URL

from aiormq.channel import Channel
from . import exceptions as exc
from .auth import AuthMechanism
from .base import Base, task
from .tools import censor_url
from .types import (
    ArgumentsType, SSLCerts,
    URLorStr)
from .version import __version__

log = logging.getLogger(__name__)


try:
    from yarl import DEFAULT_PORTS

    DEFAULT_PORTS['amqp'] = 5672
    DEFAULT_PORTS['amqps'] = 5671
except ImportError:
    pass


PRODUCT = 'aiormq'
PLATFORM = '{} {} ({} build {})'.format(
    platform.python_implementation(),
    platform.python_version(),
    *platform.python_build()
)


def parse_bool(v: str):
    return v == '1' or v.lower() in ('true', 'yes', 'y', 'enable', 'on')


def parse_int(v: str):
    try:
        return int(v)
    except ValueError:
        return 0


class Connection(Base):
    FRAME_BUFFER = 10
    # Interval between sending heartbeats based on the heartbeat(timeout)
    HEARTBEAT_INTERVAL_MULTIPLIER = 0.5
    # Allow two missed heartbeats (based on heartbeat(timeout)
    HEARTBEAT_GRACE_MULTIPLIER = 3
    _HEARTBEAT = pamqp.frame.marshal(Heartbeat(), 0)

    @staticmethod
    def _parse_ca_data(data):
        return b64decode(data) if data else data

    def __init__(self, url: URLorStr, *, parent=None,
                 loop: asyncio.AbstractEventLoop = None):

        super().__init__(
            loop=loop or asyncio.get_event_loop(),
            parent=parent
        )

        self.url = URL(url)

        if self.url.path == '/' or not self.url.path:
            self.vhost = '/'
        else:
            self.vhost = self.url.path[1:]

        self.reader = None  # type: asyncio.StreamReader
        self.writer = None  # type: asyncio.StreamWriter
        self.ssl_certs = SSLCerts(
            cafile=self.url.query.get('cafile'),
            capath=self.url.query.get('capath'),
            cadata=self._parse_ca_data(self.url.query.get('cadata')),
            key=self.url.query.get('keyfile'),
            cert=self.url.query.get('certfile'),
            verify=self.url.query.get('no_verify_ssl', '0') == '0'
        )

        self.started = False
        self.__lock = asyncio.Lock(loop=self.loop)
        self.__drain_lock = asyncio.Lock(loop=self.loop)

        self.channels = {}  # type: typing.Dict[int, typing.Optional[Channel]]

        self.server_properties = None   # type: spec.Connection.OpenOk
        self.connection_tune = None  # type: spec.Connection.TuneOk

        self.last_channel = 1

        self.heartbeat_monitoring = parse_bool(self.url.query.get(
            'heartbeat_monitoring', '1'
        ))
        self.heartbeat_timeout = parse_int(self.url.query.get(
            'heartbeat', '0'
        ))
        self.heartbeat_last_received = 0
        self.last_channel_lock = asyncio.Lock(loop=self.loop)
        self.connected = asyncio.Event(loop=self.loop)

    @property
    def lock(self):
        if self.is_closed:
            raise RuntimeError('%r closed' % self)

        return self.__lock

    async def drain(self):
        async with self.__drain_lock:
            return await self.writer.drain()

    @property
    def is_opened(self):
        return self.writer is not None and not self.is_closed

    def __str__(self):
        return str(censor_url(self.url))

    def _get_ssl_context(self):
        context = ssl.create_default_context(
            (
                ssl.Purpose.SERVER_AUTH
                if self.ssl_certs.key
                else ssl.Purpose.CLIENT_AUTH
            ),
            capath=self.ssl_certs.capath,
            cafile=self.ssl_certs.cafile,
            cadata=self.ssl_certs.cadata,
        )

        if self.ssl_certs.key:
            context.load_cert_chain(
                self.ssl_certs.cert,
                self.ssl_certs.key,
            )

        if not self.ssl_certs.verify:
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE

        return context

    @staticmethod
    def _client_capabilities():
        return {
            'platform': PLATFORM,
            'version': __version__,
            'product': PRODUCT,
            'capabilities': {
                'authentication_failure_close': True,
                'basic.nack': True,
                'connection.blocked': False,
                'consumer_cancel_notify': True,
                'publisher_confirms': True
            },
            'information': 'See https://github.com/mosquito/aiormq/',
        }

    @staticmethod
    def _credentials_class(start_frame: spec.Connection.Start):
        for mechanism in start_frame.mechanisms.decode().split():
            with suppress(KeyError):
                return AuthMechanism[mechanism]

        raise exc.AuthenticationError(
            start_frame.mechanisms,
            [m.name for m in AuthMechanism]
        )

    async def __rpc(self, request: spec.Frame, wait_response=True):
        self.writer.write(pamqp.frame.marshal(request, 0))

        if not wait_response:
            return

        _, _, frame = await self.__receive_frame()

        if request.synchronous and frame.name not in request.valid_responses:
            raise spec.AMQPInternalError(frame, frame)
        elif isinstance(frame, spec.Connection.Close):
            if frame.reply_code == 403:
                raise exc.ProbableAuthenticationError(frame.reply_text)

            raise exc.ConnectionClosed(frame.reply_code, frame.reply_text)
        return frame

    @task
    async def connect(self):
        if self.writer is not None:
            raise RuntimeError("Already connected")

        ssl_context = None

        if self.url.scheme == 'amqps':
            ssl_context = await self.loop.run_in_executor(
                None, self._get_ssl_context
            )

        try:
            self.reader, self.writer = await asyncio.open_connection(
                self.url.host, self.url.port, ssl=ssl_context,
                loop=self.loop
            )
        except OSError as e:
            raise ConnectionError(*e.args) from e

        try:
            protocol_header = ProtocolHeader()
            self.writer.write(protocol_header.marshal())

            res = await self.__receive_frame()
            _, _, frame = res   # type: spec.Connection.Start
            self.heartbeat_last_received = self.loop.time()
        except EOFError as e:
            raise exc.IncompatibleProtocolError(*e.args) from e

        credentials = self._credentials_class(frame)

        self.server_properties = frame.server_properties

        # noinspection PyTypeChecker
        self.connection_tune = await self.__rpc(spec.Connection.StartOk(
            client_properties=self._client_capabilities(),
            mechanism=credentials.name,
            response=credentials.value(self).marshal()
        ))      # type: spec.Connection.Tune

        if self.heartbeat_timeout > 0:
            self.connection_tune.heartbeat = self.heartbeat_timeout

        await self.__rpc(spec.Connection.TuneOk(
            channel_max=self.connection_tune.channel_max,
            frame_max=self.connection_tune.frame_max,
            heartbeat=self.connection_tune.heartbeat,
        ), wait_response=False)

        await self.__rpc(spec.Connection.Open(virtual_host=self.vhost))

        # noinspection PyAsyncCall
        self.create_task(self.__reader())

        # noinspection PyAsyncCall
        self.create_task(self.__heartbeat_task())
        self.loop.call_soon(self.connected.set)

        return True

    async def __heartbeat_task(self):
        if not self.connection_tune.heartbeat:
            return

        heartbeat_interval = (
            self.connection_tune.heartbeat * self.HEARTBEAT_INTERVAL_MULTIPLIER
        )
        heartbeat_grace_timeout = (
            self.connection_tune.heartbeat * self.HEARTBEAT_GRACE_MULTIPLIER
        )

        while True:
            await asyncio.sleep(heartbeat_interval, loop=self.loop)

            # Send heartbeat to server unconditionally
            self.writer.write(self._HEARTBEAT)

            if not self.heartbeat_monitoring:
                continue

            # Check if the server sent us something
            # within the heartbeat grace period
            last_heartbeat = self.loop.time() - self.heartbeat_last_received

            if last_heartbeat <= heartbeat_grace_timeout:
                continue

            await self.close(
                ConnectionError(
                    'Server connection probably hang, last heartbeat '
                    'received %.3f seconds ago' % last_heartbeat
                )
            )

            return

    @task
    async def __receive_frame(self) -> typing.Tuple[int, int, spec.Frame]:
        async with self.lock:
            frame_header = await self.reader.readexactly(1)

            if frame_header == b'\0x00':
                raise spec.AMQPFrameError(await self.reader.read())

            frame_header += await self.reader.readexactly(6)

            if not self.started and frame_header.startswith(b'AMQP'):
                raise spec.AMQPSyntaxError
            else:
                self.started = True

            frame_type, _, frame_length = pamqp.frame.frame_parts(
                frame_header
            )

            frame_payload = await self.reader.readexactly(
                frame_length + 1
            )

            return pamqp.frame.unmarshal(frame_header + frame_payload)

    @staticmethod
    def __exception_by_code(frame: spec.Connection.Close):
        if frame.reply_code == 501:
            return exc.ConnectionFrameError(frame.reply_text)
        elif frame.reply_code == 502:
            return exc.ConnectionSyntaxError(frame.reply_text)
        elif frame.reply_code == 503:
            return exc.ConnectionCommandInvalid(frame.reply_text)
        elif frame.reply_code == 504:
            return exc.ConnectionChannelError(frame.reply_text)
        elif frame.reply_code == 505:
            return exc.ConnectionUnexpectedFrame(frame.reply_text)
        elif frame.reply_code == 506:
            return exc.ConnectionResourceError(frame.reply_text)
        elif frame.reply_code == 530:
            return exc.ConnectionNotAllowed(frame.reply_text)
        elif frame.reply_code == 540:
            return exc.ConnectionNotImplemented(frame.reply_text)
        elif frame.reply_code == 541:
            return exc.ConnectionInternalError(frame.reply_text)
        else:
            return exc.ConnectionClosed(frame.reply_code, frame.reply_text)

    async def __reader(self):
        try:
            while not self.reader.at_eof():
                weight, channel, frame = await self.__receive_frame()

                self.heartbeat_last_received = self.loop.time()

                if channel == 0:
                    if isinstance(frame, spec.Connection.Close):
                        return await self.close(self.__exception_by_code(frame))
                    elif isinstance(frame, Heartbeat):
                        continue

                    log.error('Unexpected frame %r', frame)
                    continue

                if self.channels.get(channel) is None:
                    log.exception(
                        "Got frame for closed channel %d: %r", channel, frame
                    )
                    continue

                ch = self.channels[channel]

                channel_close_responses = (
                    spec.Channel.Close,
                    spec.Channel.CloseOk
                )

                if isinstance(frame, channel_close_responses):
                    self.channels[channel] = None

                await ch.frames.put((weight, frame))
        except asyncio.CancelledError as e:
            log.debug("Reader task cancelled:", exc_info=e)
        except asyncio.IncompleteReadError as e:
            log.debug("Can not read bytes from server:", exc_info=e)
            await self.close(ConnectionError(*e.args))
        except Exception as e:
            log.debug("Reader task exited because:", exc_info=e)
            await self.close(e)

    async def _on_close(self, exc=exc.ConnectionClosed(0, 'normal closed')):
        writer = self.writer
        self.reader = None
        self.writer = None

        # noinspection PyShadowingNames
        writer.close()
        return await writer.wait_closed()

    @property
    def server_capabilities(self) -> ArgumentsType:
        return self.server_properties['capabilities']

    @property
    def basic_nack(self) -> bool:
        return self.server_capabilities.get('basic.nack')

    @property
    def consumer_cancel_notify(self) -> bool:
        return self.server_capabilities.get('consumer_cancel_notify')

    @property
    def exchange_exchange_bindings(self) -> bool:
        return self.server_capabilities.get('exchange_exchange_bindings')

    @property
    def publisher_confirms(self):
        return self.server_capabilities.get('publisher_confirms')

    async def channel(self, channel_number: int = None,
                      publisher_confirms=True,
                      frame_buffer=FRAME_BUFFER, **kwargs) -> Channel:

        await self.connected.wait()

        if self.is_closed:
            raise RuntimeError('%r closed' % self)

        if not self.publisher_confirms and publisher_confirms:
            raise ValueError("Server doesn't support publisher_confirms")

        if channel_number is None:
            async with self.last_channel_lock:
                while self.last_channel in self.channels.keys():
                    self.last_channel += 1

                    if self.last_channel > 65535:
                        log.warning("Resetting channel number for %r", self)
                        self.last_channel = 1
                        # switching context for prevent blocking event-loop
                        await asyncio.sleep(0, loop=self.loop)

                channel_number = self.last_channel
        elif channel_number in self.channels:
            raise ValueError("Channel %d already used" % channel_number)

        if channel_number < 0 or channel_number > 65535:
            raise ValueError('Channel number too large')

        channel = Channel(
            self, channel_number, frame_buffer=frame_buffer,
            publisher_confirms=publisher_confirms, **kwargs
        )

        self.channels[channel_number] = channel

        try:
            await channel.open()
        except Exception:
            self.channels[channel_number] = None
            raise

        return channel

    async def __aenter__(self):
        await self.connect()


async def connect(url, *args, **kwargs) -> Connection:
    connection = Connection(url, *args, **kwargs)

    try:
        await connection.connect()
    except Exception as e:
        await connection.close(e)
        raise

    return connection



import abc
import asyncio
import typing
from contextlib import suppress
from functools import wraps
from typing import TypeVar, Type, Union

from .tools import shield


T = TypeVar('T')


class TaskWrapper:
    def __init__(self, task: asyncio.Task):
        self.task = task
        self.exception = asyncio.CancelledError

    def throw(self, exception):
        self.exception = exception
        return self.task.cancel()

    async def __inner(self):
        try:
            return await self.task
        except asyncio.CancelledError as e:
            raise self.exception from e

    def __await__(self, *args, **kwargs):
        return self.__inner().__await__()

    def cancel(self):
        return self.throw(asyncio.CancelledError)

    def __getattr__(self, item):
        return getattr(self.task, item)

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, repr(self.task))


class FutureStore:
    __slots__ = "futures", "loop", "parent"

    def __init__(self, loop: asyncio.AbstractEventLoop):
        self.futures = set()  # type: typing.Set[TaskWrapper]
        self.loop = loop  # type: asyncio.AbstractEventLoop
        self.parent = None  # type: FutureStore

    def __on_task_done(self, future):
        def remover(*_):
            nonlocal future
            if future in self.futures:
                self.futures.remove(future)

        return remover

    def add(self, future: typing.Union[asyncio.Future, TaskWrapper]):
        self.futures.add(future)
        future.add_done_callback(self.__on_task_done(future))

        if self.parent:
            self.parent.add(future)

    @shield
    async def reject_all(self, exception: Exception):
        tasks = []

        while self.futures:
            future = self.futures.pop()  # type: TaskWrapper

            if future.done():
                continue

            if isinstance(future, TaskWrapper):
                future.throw(exception or Exception)
                tasks.append(future)
            elif isinstance(future, asyncio.Future):
                future.set_exception(exception)

        if tasks:
            await asyncio.gather(
                *tasks, loop=self.loop,
                return_exceptions=True
            )

    def create_task(self, coro: T) -> T:
        task = TaskWrapper(self.loop.create_task(coro))
        self.add(task)
        return task

    def create_future(self):
        future = self.loop.create_future()
        self.add(future)
        return future

    def get_child(self) -> "FutureStore":
        store = FutureStore(self.loop)
        store.parent = self
        return store


class Base:
    __slots__ = 'loop', '__future_store', 'closing'

    def __init__(self, *, loop, parent: 'Base' = None):
        self.loop = loop      # type: asyncio.AbstractEventLoop

        if parent:
            self.__future_store = parent._future_store_child()
        else:
            self.__future_store = FutureStore(loop=self.loop)

        self.closing = self._create_closing_future()

    def _create_closing_future(self):
        future = self.__future_store.create_future()
        future.add_done_callback(lambda x: x.exception())
        return future

    def _cancel_tasks(self, exc: Union[Exception, Type[Exception]] = None):
        return self.__future_store.reject_all(exc)

    def _future_store_child(self):
        return self.__future_store.get_child()

    # noinspection PyShadowingNames
    def create_task(self, coro) -> asyncio.Future:
        return self.__future_store.create_task(coro)

    def create_future(self) -> asyncio.Future:
        return self.__future_store.create_future()

    @abc.abstractmethod
    async def _on_close(self, exc=None):  # pragma: no cover
        return

    async def __closer(self, exc):
        if self.is_closed:  # pragma: no cover
            return

        with suppress(Exception):
            await self._on_close(exc)

        with suppress(Exception):
            await self._cancel_tasks(exc)

    async def close(self, exc=asyncio.CancelledError()):
        if self.is_closed:
            return

        await self.loop.create_task(self.__closer(exc))

    def __repr__(self):
        cls_name = self.__class__.__name__
        return '<{0}: "{1}">'.format(cls_name, str(self))

    @abc.abstractmethod
    def __str__(self):  # pragma: no cover
        raise NotImplementedError

    @property
    def is_closed(self):
        return self.closing.done()


def task(func: T) -> T:
    @wraps(func)
    async def wrap(self: "Base", *args, **kwargs):
        # noinspection PyCallingNonCallable
        return await self.create_task(func(self, *args, **kwargs))

    return wrap
